<?php
/**
 * XModule - Xpell Base Module
 * This class is being extended by modules with the following logic:
 * 1. Every module must have a name
 * 2. Module holds Object manager to manage the module specific object (extends XObject)
 * 3. Every module can execute XCommand (XCommand, JSON, text(CLI style)),
 *    the rules of the method invocation is the "underscore" sign, meaning functions that will start with "_" sign
 *    will be exposed to XPell Interpreter
 *
 * @example
 *  '''
 *    The following module:
 *      
 *    class myModule extends XModule {
 *          public function __construct() {...}
 *          public function _my_Command($xCommand) {
 *              ...
 *          }
 *    }
 * 
 *    will be called like this:
 * 
 *    XModule::execute("my-Command")
 *      - when calling the method there is no need for the underscore sign, spaces and dashes will be converted to underscore
 *  '''
 */

require_once 'XUtils.php';
require_once 'XParser.php';
require_once 'XLogger.php';
require_once 'XObjectManager.php';
require_once 'XObject.php';
require_once 'XCommand.php';


class XModule {
    protected $_id;
    public $_name;
    protected $_loaded = false;
    protected $_log_rules = [
        'load' => true,
        'createObject' => false,
        'removeObject' => false,
    ];
    protected $_object_manager;

    public function __construct($data) {
        $this->_name = $data['_name'];
        $this->_id = XUtils::guid();
        $this->_object_manager = new XObjectManager($this->_name);
    }

    public function load() {
        if ($this->_log_rules['load']) {
            // XLogger::log("Module " . $this->_name . " loaded");
        }
        $this->_loaded = true;
        // add the module to xpell main interpreter
    }

    public function create($data) {
        $xObject = null;
        if (isset($data['_type'])) {
            if ($this->_object_manager->hasObjectClass($data['_type'])) {
                $xObjectClass = $this->_object_manager->getObjectClass($data['_type']);
                if (isset($xObjectClass->defaults)) {
                    XUtils::mergeDefaultsWithData($data, $xObjectClass->defaults);
                }
                $xObject = new $xObjectClass($data);
            } else {
                throw new Exception("Xpell object '" . $data['_type'] . "' not found");
            }
        } else {
            $xObject = new XObject($data);
        }

        $this->_object_manager->addObject($xObject);
        if (isset($data['_children'])) {
            foreach ($data['_children'] as $spell) {
                $new_spell = $this->create($spell);
                $xObject->append($new_spell);
            }
        }
        $xObject->onCreate();
        return $xObject;
    }

    public function remove($objectId) {
        $obj = $this->_object_manager->getObject($objectId);
        if ($obj) {
            $this->_object_manager->removeObject($objectId);
            if (method_exists($obj, 'dispose')) {
                $obj->dispose();
            }
        }
    }

    public function _info($xCommand) {
        XLogger::log("module info");
    }

    public function run($stringXCommand) {
        if ($stringXCommand) {
            $strCmd = trim($stringXCommand);
            if (strpos($strCmd, $this->_name) !== 0) {
                $strCmd = $this->_name . " " . $strCmd;
            }
            $xCommand = XParser::parse($strCmd);
            return $this->execute($xCommand);
        } else {
            throw new Exception("Unable to parse Xpell Command");
        }
    }

    public function execute($xCommand) {
        if (isset($xCommand->_op)) {
            $lop = "_" . str_replace('-', '_', $xCommand->_op);
            if (method_exists($this, $lop)) {
                return $this->$lop($xCommand);
            } else if ($this->_object_manager) {
                $o = $this->_object_manager->getObjectByName($xCommand->_op);
                if ($o) {
                    $o->execute($xCommand);
                } else {
                    throw new Exception("Xpell Module can't find op: " . $xCommand->_op);
                }
            } else {
                throw new Exception("Xpell Module can't find op: " . $xCommand->_op);
            }
        }
    }

    public function onFrame($frameNumber) {
        $omObjects = $this->_object_manager->get_objects();
        foreach ($omObjects as $key => $onFrameCallBack) {
            if (method_exists($onFrameCallBack, 'onFrame')) {
                $onFrameCallBack->onFrame($frameNumber);
            }
        }
    }

    public function getOm() {
        return $this->_object_manager;
    }

    public function _o($objectId) {
        return $this->getObject($objectId);
    }

    public function getObject($objectId) {
        return $this->_object_manager->getObject($objectId);
    }

    public function importObjectPack($xObjectPack) {
        $this->_object_manager->registerObjects($xObjectPack->getObjects());
    }

    public function importObjects($xObjectPack) {
        $this->importObjectPack($xObjectPack);
    }

    public function importObject($xObjectName, $xObject) {
        $this->_object_manager->registerObject($xObjectName, $xObject);
    }
}

?>