<?php
/**
 * XObject - Base Xpell object for Xpell modules
 */

require_once 'XUtils.php';
require_once 'XCommand.php';
require_once 'XParser.php';
require_once 'XLogger.php';
require_once 'XEventManager.php';
require_once 'XData.php';

interface IXData {
    public function __get($key);
    public function __set($key, $value);
}



class XObject {
    public $_id;
    public $_type;
    public $_children = [];
    public $_name;
    public $_data_source;
    public $_on = [];
    public $_once = [];
    public $_on_create;
    public $_on_mount;
    public $_on_frame;
    public $_on_data;
    public $_on_event;
    public $_process_frame = true;
    public $_process_data = true;
    protected $_xem_options;
    protected $_nano_commands = [];
    protected $_cache_cmd_txt;
    protected $_cache_jcmd;
    protected $_event_listeners_ids = [];
    protected $_xporter = [
        '_ignore_fields' => ["_to_xdata_ignore_fields", "_xporter", "_children", "_on", "_once", "_on_create", "_on_mount", "_on_frame", "_on_data", "_process_frame", "_process_data", "_xem_options", "_event_listeners_ids"],
        '_instance_xporters' => []
    ];

    public function __construct($data, $defaults = null, $skipParse = false) {
        if ($defaults) {
            XUtils::mergeDefaultsWithData($data, $defaults);
        }

        $this->_id = isset($data['_id']) ? $data['_id'] : "xo-" . XUtils::guid();
        $this->_type = "object";
        $this->_children = [];
        $this->_nano_commands = [];
        // $this->addNanoCommandPack([
        //     "info" => function($xCommand, $xObject = null) {
        //         XLogger::log("XObject id " . ($xObject ? $xObject->_id : ""));
        //     },
        //     "log" => function($xCommand, $xObject = null) {
        //         if (isset($xCommand->_params) && isset($xCommand->_params["1"])) {
        //             XLogger::log($xCommand->_params["1"]);
        //         } else {
        //             XLogger::log($xObject);
        //         }
        //     },
        //     "fire" => function($xCommand, $xObject = null) {
        //         if (isset($xCommand->_params) && isset($xCommand->_params["1"])) {
        //             XEventManager::fire($xCommand->_params["1"]);
        //         } else if (isset($xCommand->_params) && isset($xCommand->_params["event"])) {
        //             XEventManager::fire($xCommand->_params["event"]);
        //         }
        //     }
        // ]);

        // $this->addXporterDataIgnoreFields(["_nano_commands"]);
        // $this->addXporterInstanceXporter(XObject::class, function($objectInstance) {
        //     return $objectInstance->toXData();
        // });


        $this->init($data, $skipParse);
    }

    public function init($data = null, $skipParse = false) {
        if (!$skipParse && $data) {
            unset($data['_id']);
            $this->parse($data, ['_children' => "child nodes"]);
            $this->parseEvents();
        }
    }

    public function parseEvents($options = null) {
        foreach ($this->_on as $eventName => $handler) {
            if (is_callable($handler)) {
                $this->addEventListener($eventName, $handler, $options);
            } else {
                throw new Exception("event handler must be a function " . $eventName);
            }
        }

        $onceOptions = $options;
        $onceOptions['_once'] = true;

        foreach ($this->_once as $eventName => $handler) {
            if (is_callable($handler)) {
                $this->addEventListener($eventName, $handler, $onceOptions);
            } else {
                throw new Exception("event handler must be a function");
            }
        }
    }

    public function addEventListener($eventName, $handler, $options = null) {
        $xem = XEventManager::class;
        $event_listener_id = $xem::on($eventName, function($eventData) use ($handler) {
            $handler($this, $eventData);
        }, $options);
        $this->_event_listeners_ids[$eventName] = $event_listener_id;
    }

    public function removeEventListener($eventName) {
        if (isset($this->_event_listeners_ids[$eventName])) {
            XEventManager::remove($this->_event_listeners_ids[$eventName]);
            unset($this->_event_listeners_ids[$eventName]);
        }
    }

    public function removeAllEventListeners() {
        foreach ($this->_event_listeners_ids as $key => $value) {
            $this->removeEventListener($key);
        }
    }

    public function append($xobject) {
        //push to children
        $this->_children[] = $xobject; 
    }

    public function remove($xobject) {
        $index = array_search($xobject, $this->_children);
        if ($index !== false) {
            array_splice($this->_children, $index, 1);
        }
    }

    public function addNanoCommand($commandName, $nanoCommandFunction) {
        if (is_callable($nanoCommandFunction)) {
            $this->_nano_commands[$commandName] = $nanoCommandFunction;
        }
    }

    public function addNanoCommandPack($ncPack) {
        if ($ncPack) {
            foreach ($ncPack as $key => $value) {
                $this->addNanoCommand($key, $value);
            }
        }
    }

    // public function addXporterDataIgnoreFields($ignoreFields) {
    //     $this->_xporter['_ignore_fields'] = array_merge($this->_xporter['_ignore_fields'], $ignoreFields);
    // }

    public function addXporterInstanceXporter($classOfInstance, $handler) {
        $xporterName = XUtils::guid();
        $this->_xporter['_instance_xporters'][$xporterName] = [
            'cls' => $classOfInstance,
            'handler' => $handler
        ];
    }

    public function dispose() {
        $this->_process_data = false;
        $this->_process_frame = false;
        $this->removeAllEventListeners();
        foreach ($this->_children as $child) {
            if (method_exists($child, 'dispose')) {
                $child->dispose();
            }
        }
        $this->_children = [];
    }

    public function parse($data, $ignore = ['_children' => "child nodes"]) {
        foreach ($data as $field => $value) {
            if (!array_key_exists($field, $ignore) && array_key_exists($field, $data)) {
                $this->$field = $value;
            }
        }
    }

    public function parseFieldsFromXDataObject($data, $fields) {
        foreach ($fields as $field => $default) {
            if (array_key_exists($field, $data)) {
                $this->$field = $data[$field];
            } else {
                $this->$field = $default;
            }
        }
    }

    public function parseFields($data, $fields, $checkNonXParams = false) {
        foreach ($fields as $field) {
            if (array_key_exists($field, $data)) {
                $this->$field = $data[$field];
            } else if ($checkNonXParams && strpos($field, '_') === 0) {
                $choppedField = substr($field, 1);
                if (array_key_exists($choppedField, $data)) {
                    $this->$field = $data[$choppedField];
                    $this->$choppedField = $data[$choppedField];
                }
            }
        }
    }

    public function onCreate() {
        if ($this->_on_create) {
            if (is_callable($this->_on_create)) {
                call_user_func($this->_on_create, $this);
            } else if (is_string($this->_on_create)) {
                $this->run($this->_id . " " . $this->_on_create);
            }
        }
        foreach ($this->_children as $child) {
            if (method_exists($child, 'onCreate')) {
                $child->onCreate();
            }
        }
    }

    public function onMount() {
        if ($this->_on_mount) {
            if (is_callable($this->_on_mount)) {
                call_user_func($this->_on_mount, $this);
            } else if (is_string($this->_on_mount)) {
                $this->run($this->_id . " " . $this->_on_mount);
            }
        }
        foreach ($this->_children as $child) {
            if (method_exists($child, 'onMount')) {
                $child->onMount();
            }
        }
    }

    public function emptyDataSource() {
        if ($this->_data_source && is_string($this->_data_source)) {
            XData::delete($this->_data_source);
        }
    }

    public function onData($data) {
        if ($this->_on_data && $this->_process_data) {
            if (is_callable($this->_on_data)) {
                call_user_func($this->_on_data, $this, $data);
            } else if (is_string($this->_on_data)) {
                $this->run($this->_id . " " . $this->_on_data);
            }
        }
    }

    public function onFrame($frameNumber) {
        if ($this->_on_frame && $this->_process_frame) {
            if (is_callable($this->_on_frame)) {
                if (is_callable($this->_on_frame)) {
                    call_user_func($this->_on_frame, $this, $frameNumber);
                }
            } else if (is_string($this->_on_frame)) {
                $this->run($this->_id . " " . $this->_on_frame);
            }
        }

        if ($this->_data_source && $this->_process_data) {
            if (XData::has($this->_data_source)) {
                $this->onData(XData::$objects[$this->_data_source]);
            }
        }

        foreach ($this->_children as $child) {
            if (method_exists($child, 'onFrame')) {
                $child->onFrame($frameNumber);
            }
        }
    }

    public function run($nanoCommand, $cache = true) {
        $jcmd = ($this->_cache_cmd_txt && $this->_cache_cmd_txt == $nanoCommand) ? $this->_cache_jcmd : XParser::parse($nanoCommand);
        if ($cache) {
            $this->_cache_cmd_txt = $nanoCommand;
            $this->_cache_jcmd = $jcmd;
        }
        $this->execute($jcmd);
    }

    public function execute($xCommand) {
        if (isset($xCommand->_op) && isset($this->_nano_commands[$xCommand->_op])) {
            try {
                $this->_nano_commands[$xCommand->_op]($xCommand, $this);
            } catch (Exception $err) {
                XLogger::error($this->_id . " has error with command name " . $xCommand->_op . " " . $err->getMessage());
            }
        } else {
            XLogger::error($this->_id . " has no command name " . $xCommand->_op);
        }
    }

    public function toXData() {
        $out = [];
        foreach ($this as $field => $value) {
            if (!in_array($field, $this->_xporter['_ignore_fields']) && isset($this->$field)) {
                if (is_callable($value)) {
                    $funcStr = (new ReflectionFunction($value))->getName();
                    if (strpos($funcStr, 'class') !== 0) {
                        $out[$field] = $funcStr;
                    }
                } else if (is_object($value)) {
                    $xporters = array_keys($this->_xporter['_instance_xporters']);
                    $regField = true;
                    foreach ($xporters as $xporter) {
                        $xp = $this->_xporter['_instance_xporters'][$xporter];
                        if ($value instanceof $xp['cls']) {
                            $out[$field] = $xp['handler']($value);
                            $regField = false;
                        }
                    }
                    if ($regField) {
                        $out[$field] = $value;
                    }
                } else {
                    $out[$field] = $value;
                }
            }
        }
        $out['_children'] = [];
        if (count($this->_children) > 0) {
            foreach ($this->_children as $child) {
                if (method_exists($child, 'toXData')) {
                    $out['_children'][] = $child->toXData();
                }
            }
        }
        return $out;
    }

    public function __toString() {
        return json_encode($this->toXData());
    }
}

class XObjectPack {
    public static function getObjects() {
        return [
            "object" => XObject::class
        ];
    }
}

?>