<?php

require_once 'XCommand.php';

/**
 * Xpell Parser - Parse XML, HTML, Raw Text & Json to Xpell Command
 */
class XParser {

    private static $html2XMap = [
        'elements' => [
            'div' => 'view',
            'a' => 'link',
            'b' => 'xhtml',
            'h1' => 'xhtml', 'h2' => 'xhtml', 'h3' => 'xhtml', 'h4' => 'xhtml', 'h5' => 'xhtml', 'p' => 'xhtml', 'small' => 'xhtml', 'aside' => 'xhtml', 'span' => 'xhtml',
            'table' => 'xhtml', 'th' => 'xhtml', 'td' => 'xhtml', 'tr' => 'xhtml', 'thead' => 'xhtml', 'tbody' => 'xhtml',
            'ul' => 'xhtml', 'li' => 'xhtml', 'ol' => 'xhtml',
            'canvas' => 'xhtml',
            'img' => 'image',
        ],
        'attributes' => [
            'id' => '_id'
        ]
    ];

    /**
     * Adds HTML-Xpell Mapping item
     * @param string $htmlElement HTML element to change from
     * @param string $xpellElement Xpell element to change to 
     */
    public static function addHtml2XpellMapItem($htmlElement, $xpellElement) {
        self::$html2XMap['elements'][$htmlElement] = $xpellElement;
    }

    /**
     * Convert text command to Xpell json command
     * @param string $txt
     * @param string|null $module
     * @return XCommand
     */
    public static function parse($txt, $module = null) {
        $carr = explode(' ', $txt);
        $rv = new XCommand();
        if ($module) {
            $rv->_module = $module;
            $rv->_op = $carr[0];
        } else {
            $rv->_module = $carr[0];
            $rv->_op = $carr[1];
        }
        $rv->_params = [];

        if (count($carr) > 1) {
            for ($i = 2; $i < count($carr); ++$i) {
                $v = $carr[$i];
                $dl = strpos($v, ':');
                if ($dl > -1) {
                    $mc = explode(':', $v);
                    $rv->_params[$mc[0]] = $mc[1];
                } else {
                    $rv->_params[$i - 1] = $carr[$i];
                }
            }
        }

        return $rv;
    }

    public static function replaceSpacesInQuotes($inputString, $replaceWith = '_%20_') {
        return preg_replace_callback('/([\'"])(.*?)\1/', function ($matches) use ($replaceWith) {
            return str_replace(' ', $replaceWith, $matches[2]);
        }, $inputString);
    }

    public static function parseXpellCommand($command, $module = null) {
        $command = self::replaceSpacesInQuotes($command);
        $parts = preg_split('/\s+/', trim($command));

        $moduleName = $module ? $module : array_shift($parts);
        $op = array_shift($parts);

        $params = [];
        $currentParam = null;
        $valueInProgress = null;

        foreach ($parts as $part) {
            if ($valueInProgress) {
                $valueInProgress .= " $part";
                if (substr($part, -1) === substr($valueInProgress, 0, 1)) {
                    $value = substr($valueInProgress, 1, -1);
                    $params[$currentParam] = str_replace('_%20_', ' ', $value);
                    $valueInProgress = null;
                }
            } elseif (preg_match('/^["\']/', $part)) {
                $quoteChar = $part[0];
                $valueInProgress = $part;
                if (substr($part, -1) === $quoteChar) {
                    $value = substr($part, 1, -1);
                    $params[$currentParam] = str_replace('_%20_', ' ', $value);
                    $valueInProgress = null;
                }
            } elseif (strpos($part, ':') !== false) {
                list($param, $value) = explode(':', $part, 2);
                $params[$param] = str_replace('_%20_', ' ', $value);
                if (strpos($value, ' ') !== false) {
                    throw new Exception("Parameter value '$value' with spaces must be enclosed in double or single quotes.");
                }
            } else {
                if ($currentParam) {
                    $params[$currentParam] .= " $part";
                }
            }

            if (!$valueInProgress && !$currentParam && strpos($part, ':') === false) {
                $currentParam = str_replace('_%20_', ' ', $part);
            }
        }

        return [
            '_module' => $moduleName,
            '_op' => $op,
            '_params' => $params
        ];
    }

    /**
     * Covent Xpell2 (XP2) Json to Xpell JSON
     * @param array $XP2Json Xpell 2 JSON
     * @return array
     */
    public static function xpellify($XP2Json) {
        $tkeys = array_keys($XP2Json);
        $outputXpell = ['_type' => $tkeys[0]];
        $outputXpell['_children'] = [];
        $firstObject = $XP2Json[$tkeys[0]];
        $foKeys = array_keys($firstObject);

        foreach ($foKeys as $iKey) {
            if ($iKey === '_attr') {
                $outputXpell = array_merge($outputXpell, $firstObject[$iKey]);
            } else {
                $lob = [];
                $lob[$iKey] = $firstObject[$iKey];
                $outputXpell['_children'][] = self::xpellify($lob);
            }
        }
        return $outputXpell;
    }

    /**
     * Converts XML/HTML string to XCommand
     * @param string $xmlString XML string
     * @return array
     */
    public static function xmlString2Xpell($xmlString) {
        $xmlDoc = new DOMDocument();
        $xmlDoc->loadXML($xmlString);
        if ($xmlDoc->childNodes->length > 0) {
            return self::xml2Xpell($xmlDoc->childNodes->item(0));
        } else {
            return [];
        }
    }

    /**
     * Converts XML/HTML Document to Xpell JSON
     * @param DOMNode $xmlNode XML Document Node
     * @param bool $forceXhtml force Xpell XHTML for every unknown object
     * @return array Xpell JSON
     */
    public static function xml2Xpell($xmlNode, $forceXhtml = false) {
        $cMap = self::$html2XMap;
        $scanChildren = true;
        $outputXpell = [];
        $outputXpell['_children'] = [];
        $root_name = $xmlNode->nodeName;
        $html_tag_attr = $xmlNode->nodeName;
        $forceXhtmlOnChildren = $forceXhtml;

        if ($forceXhtml) {
            $outputXpell['_type'] = 'xhtml';
            $outputXpell['_html_ns'] = 'http://www.w3.org/2000/svg';
        } else {
            $outputXpell['_type'] = isset($cMap['elements'][$root_name]) ? $cMap['elements'][$root_name] : $root_name;
        }

        if ($xmlNode->attributes) {
            foreach ($xmlNode->attributes as $n) {
                $attr_name = isset($cMap['attributes'][$n->name]) ? $cMap['attributes'][$n->name] : $n->name;
                $outputXpell[$attr_name] = $n->value;
            }
        }

        if ($xmlNode->firstChild && $xmlNode->firstChild->nodeValue) {
            $outputXpell['text'] = trim($xmlNode->firstChild->nodeValue);
        }

        if ($outputXpell['_type'] == 'xhtml') {
            $outputXpell['_html_tag'] = $html_tag_attr;
        } elseif ($outputXpell['_type'] == 'svg') {
            $forceXhtmlOnChildren = true;
            $outputXpell['_html_ns'] = 'http://www.w3.org/2000/svg';
        }

        if ($scanChildren && $xmlNode->childNodes->length > 0) {
            foreach ($xmlNode->childNodes as $node) {
                if (strpos($node->nodeName, '#') !== 0) {
                    $outputXpell['_children'][] = self::xml2Xpell($node, $forceXhtmlOnChildren);
                }
            }
        }

        return $outputXpell;
    }
}

?>